#include <stdio.h>
#include <math.h>
#include <S.h>
#if defined(SPLUS_VERSION) && SPLUS_VERSION >= 4000
#  include <newredef.h>
#endif

/* Adjusted prediction error G(M) for the WARPing-              */
/* approximation of the NADARAYA-WATSON estimate                */
double selector1 (u,v)          /* Shibata's model selector     */
double u,v;
/* u <==> K(0), v <==> fM */
{
  double w;
  w=1+2*u/v;
  return(w);
}
double selector2 (u,v)          /* Generalised Cross-Validation */
double u,v;
{
  double w;
  w=pow(1.0-u/v,-2.0);
  return(w);
}
double selector3 (u,v)          /* Akaike's Information Criterion */
double u,v;
{
  double w;
  w=exp(2*u/v);
  return(w);
}
double selector4 (u,v)          /* Finite Prediction Error      */
double u,v;
{
  double w;
  w=(1.0+u/v)/(1.0-u/v);
  return(w);
}
double selector5 (u,v)          /* Rice's T                     */
double u,v;
{
  double w;
  w=1.0/(1.0-2.0*u/v);
  return(w);
}

Gwarpingreg(x,y,delta,rangeM,score,kwe,origin,bin,counts,ysum,ysquaresum,
index,n,numbin,kernel,select,binboundary,indexweight)
/* x <==> 1 dim predictor                                       */
/* y <==> 1 dim response                                        */
/* delta <==> binwidth                                          */
/* rangeM <==> range of smoothing parameter M                   */
/* score <==> value of the score function                       */
/* kwe <==> weights depending on the different kernels          */
/* origin <==> left boundary of first bin                       */
/* bin <==> flag for non-empty bins                             */
/* counts <==> counts in non-empty bins                         */
/* ysum <==> sum of response-observations in non-empty bins     */
/* ysquaresum <==> sum of squared response in non-empty bins    */
/* index <==> index of non-empty bins                           */
/* n <==> number of observations                                */
/* numbin <==> total number of small bins                       */
/* kernel <==>  type of kernel coded from 1 to 5                */
/* select <==>   type of selection function coded 1 to 5        */
/* binboundary <==> index of first and last bin with weight 1   */
/* indexweight <==> index of non-empty bins with weight 1       */
double *x,*y,*delta,*score,*kwe,*origin,*ysum,*ysquaresum;
long *rangeM,*bin,*counts,*index,*n,*numbin,*kernel,*select,
	*binboundary,*indexweight;
{
  double mM,rM,fM,term1,term2;
  long index2,nl,nlweight;
  register long M,i,j,k;
  double(*selector)();
  nl=0;
  nlweight=0;
  switch ( *select)
  {
    case 1: selector=selector1; break;
    case 2: selector=selector2; break;
    case 3: selector=selector3; break;
    case 4: selector=selector4; break;
    case 5: selector=selector5; break;
    default:
    printf("Error in the choice of selector !");return;
  }
/* BINNING THE DATA */
  for (i=0;i<n[0];i++)
  {
    j=floor((x[i]-origin[0])/(*delta));
    if (bin[j]==0) /* first observ. in bin */
    {
      ++nl;
      counts[nl]=1;
      ysum[nl]=y[i];
      ysquaresum[nl]=pow(y[i],2.0);
      index[nl]=j;
      bin[j]=nl;
      if ((j>=binboundary[0]) && ( j<=binboundary[1]))
      {
	++nlweight;
	indexweight[nlweight]=nl;
      }
    }
    else
    {
      counts[bin[j]]++;
      ysum[bin[j]]+=y[i];
      ysquaresum[bin[j]]+=pow(y[i],2.0);
    }
  }
  for (M=rangeM[0];M<(rangeM[1]+1);M++)
  {
/* CREATING WEIGHTS */
/* NOTE WEIGHTS ARE NOT NORMALISED */
    switch ( *kernel )
    {
      case 1 : for (i=0;i<M;i++)
	kwe[i]=1;
	break;
      case 2 : for (i=0;i<M;i++)
	kwe[i]=1.0-(double)i/(double)M;
	break;
      case 3 : for (i=0;i<M;i++)
	kwe[i]=1.0-pow((double)i/(double)M, 2.0);
	break;
      case 4 : for (i=0;i<M;i++)
	kwe[i]=pow(1.0-pow((double)i/(double)M,2.0),2.0);
	break;
      case 5 : for (i=0;i<M;i++)
	kwe[i]=pow(1.0-pow((double)i/(double)M,2.0),3.0);
	break;
      default: 
	printf("error in the choice of kernel %d!", *kernel);
	return;
    }
    for (j=1;j<nlweight+1;j++)
    {
 /* these were never set, so the first M value was rubbish */
      rM=0;
      fM=0;
      k=indexweight[j];
      for (i=1-M;i<M;i++)
      {
	index2=bin[index[k]+i];
	if(index2>0)
	{
	  rM += kwe[abs(i)]*ysum[index2];
	  fM += kwe[abs(i)]*counts[index2];
	}
      }
      mM=rM/fM;
      term1=ysquaresum[k]-2*mM*ysum[k]+counts[k]*pow(mM,2.0);
      term2=(*selector)(kwe[0],fM);
      score[M-rangeM[0]]+=term1*term2;
   /* Initialize moved to top of loop */
    } /* end for j */
  } /* end for M */
} /* end*/
/* Algorithm for calculation of the Biased-Cross-Validation     */
/* for WARPing density estimation                               */

cvbwarping(x,fm,delta,rangem,cv,kwe,origin,bin,frequ,index,n,numbin,kernel)

/* x <==> data                                                  */
/* fm <==> vector == 0 will contain density estimates           */
/* delta bandwidth                                              */
/* rangem <==> range of smoothing parameter M                   */
/* cv <==> value of cross-validation corresponding to M         */
/* kwe <==> weights depending on the different kernels          */
/* origin <==> left border of first bin                         */
/* bin <==> vector == 0 protocol for non-empty bins             */
/* frequ <==> absolute frequencies of non-empty bins            */
/* index <==> index of non-empty bins                           */
/* n <==> number of data                                        */
/* numbin <==> number of small bins                             */
/* kernel <==> kernel=1 <==> Quartic, =2 <==> Triweight         */

double*x,*fm,*delta,*cv,*kwe,*origin;
long*rangem,*n,*kernel,*bin,*frequ,*index,*numbin;
{
  register long M,i,j,k,indexi,z;
  double zeta,kappa,tau,divisor,cm,h,score;
  double m[9];
  long nl,iabs;
  nl=0;
  
/* Binning the data                                             */
/* Computation of index[] and frequ[] for the non-empty         */
/* bins.  Number of last non-empty bin is nl                    */
 
  for (i=0;i<n[0];i++)
  {
    indexi=floor((x[i]-origin[0])/delta[0]);
/* the actual obs. belongs to the bin with number indexi                */
    if (bin[indexi]==0)
/* if the actual observation is the first in that bin           */
    {
      nl++;
/* the number of non-empty bins increases by 1                  */
      frequ[nl] = 1;
/* the number of obs. in this bin is 1 (at this time)           */
      index[nl]=indexi;
      bin[indexi]=nl;
/* protocol of the position of the new                          */
/* non-empty bin (in frequ[] and index[])                       */
    }
    else
/* if the actual observation is not the first in this bin       */
    {
      frequ[bin[indexi]]++;
/* increase the number of obs. in belonging non-empty bin       */
    }
  }
/* The vector frequ contains nonzero values in positions        */
/* 1,2,3,..,nl.  The other values are prespecified      */
/* as 0, since frequ is installed as a 0-vector.  In            */
/* analogy index contains the indices of the non-empty bins.    */
  for(M=rangem[0];M<rangem[1]+1;M++)
  {
    m[1]=(double)M;
/* Computation of first 8 powers of M in double m[i]=M^i        */
    for(i=2;i<9;i++)
      m[i] = pow(m[1], (double)i);
    h=(double)M*delta[0];
/* resetting the vector for the density estimates               */
    for(z=0;z<numbin[0];z++)
      fm[z]=0.0;
/* resetting the score                                          */
    score=0.0;
/* Calculation of the functionals of M and kernel K             */

    switch ( *kernel )
    {
      case 1 :
      { /* Quartic */
	tau=(16*m[4] +32*m[2]-13)/(112*m[2]+28);
	divisor=(16*m[4]-1)*(4*m[2]+1);
	kappa=15*m[1]/7*(1-(32*m[4]-28*m[2]+1)/divisor);
	zeta=30/7*(336*m[3]-420*m[2]+85*m[1]+104)/divisor;
	break;
      }
      case 2 :
      { /* Triweight */
	tau=(32*m[6]+80*m[4] +62*m[2]-69)/(288*m[4]+72*m[2]+60);
	divisor=(4*m[2]-1)*pow(24*m[4]+6*m[2]+5,2.0);
	kappa=350*m[1]/143*(1-(4492.8*m[8]-979.2*m[6]+2124*m[4]
		  -866.4*m[2]-2755.2)/(3*divisor));
	zeta=630*(128*m[7]-448*m[5]+560*m[4]-280*m[3]-280*m[2]
			+320*m[1]+35)/divisor;
	break;
      }
      default: printf("error in the choice of kernel !");
      return;
    }
/* end of the switch-part                                       */

/* Calculation of the kernel depending weights.                 */
/* The factor c(M) assures that sum(w.M(i/M),1-M,M-1)=M.        */
/* Note, in formula of cm the division by [n * h] does not      */
/* belong to the theoretical value of cm, but is a factor,      */
/* which has to be multiplied one time, since fm=1/(nh)*..      */

    switch ( *kernel )
    {
      case 1 :
/* Quartic - kernel K(u)= c(M) * (1- u*u)^2 * I(|u|<=1)         */
/* c(M)=15*M^4/(16*M^4-1)                                       */
      {
	cm=0.9375/((1.0-0.0625*pow((double)M,-4.0))
					*(double)n[0]*h);
	for (i=0;i<M;i++)
	  kwe[i]=cm*pow(1-pow((double)i/(double)M,2.0),2.0);
	break;
      }
      case 2 :
/* Triweight-kernel K(u)=c(M)*(1-u*U)^3*I(|u|<=1)               */
/* c(M)=35*M^6/(32*M^6+14/3*M^2-5/3)                            */
      {
	cm=1.09375/((1.0+0.14583333*pow((double)M,-4.0)
	-0.052083333*pow((double)M,-6.0))*(double)n[0]*h);
	for (i=0;i<M;i++)
	  kwe[i]=cm*pow(1-pow((double)i/(double)M,2.0),3.0);
	break;
      }
      default: printf("error in the choice of kernel !");
      return;
    }
/* end of the switch-part                                       */
/* Like Kernel estimates, where a kernel function is set        */
/* around the observations, we group the calculated weights     */
/* around the non-empty bins.                                   */

    for (k=1;k<nl+1;k++)
    {
      for (i=1-M;i<M;i++)
      {
	fm[i+index[k]]=fm[i+index[k]]+kwe[abs(i)]*(double)frequ[k];
      }
    }
/* Calculation of sum over second differences                   */
    for(z=1;z<numbin[0]-1;z++)
      score+=pow(fm[z+1]-2*fm[z]+fm[z-1],2.0);
    score+=pow(2*fm[numbin[0]-1]-fm[numbin[0]-2],2.0);
    score+=pow(2*fm[0]-fm[1],2.0);
    score+=pow(fm[0],2.0)+pow(fm[numbin[0]-1],2.0);
    score*=delta[0]*0.25;
    score-=zeta/(4*(double)n[0]*m[2]*delta[0]);
    score*=pow(tau,2.0)+1/180;  
    cv[M-rangem[0]]=score+kappa/(3*(double)n[0]*m[2]*delta[0]);
  }  /* end of loop over M */
    
}/* end */
   
					
/* Algorithm for calculation of the L2-Cross-Validation         */
/* for WARPing density estimation                               */

cvwarping(x,fm,delta,rangem,cv,kwe,origin,bin,frequ,index,n,numbin,kernel)

/* x <==> data                                                  */
/* fm <==> vector == 0 will contain density estimates           */
/* delta bandwidth                                              */
/* rangem <==> range of smoothing parameter M                   */
/* cv <==> value of cross-validation corresponding to M         */
/* kwe <==> weights depending on the different kernels          */
/* origin <==> left border of first bin                         */
/* bin <==> vector == 0 protocol for non-empty bins             */
/* frequ <==> absolute frequencies of non-empty bins            */
/* index <==> index of non-empty bins                           */
/* n <==> number of data                                        */
/* numbin <==> number of small bins                             */
/* kernel <==> type of kernel coded with 1 to 5                 */

double*x,*fm,*delta,*cv,*kwe,*origin;
long*rangem,*n,*kernel,*bin,*frequ,*index,*numbin;
{
  register long M,i,j,k,indexi,z;
  double cm,h,estexp,intsquare;
  long nl,iabs;
  nl=0;
  
/* Binning the data                                             */
/* Computation of index[] and frequ[] for the non-empty         */
/* bins.  Number of last non-empty bin is nl                    */
 
  for (i=0;i<n[0];i++)
  {
    indexi=floor((x[i]-origin[0])/delta[0]);
/* the actual obs. belongs to the bin with number indexi                */
    if (bin[indexi]==0)
/* if the actual observation is the first in that bin           */
    {
      nl++;
/* the number of non-empty bins increases by 1                  */
      frequ[nl] = 1;
/* the number of obs. in this bin is 1 (at this time)           */
      index[nl]=indexi;
      bin[indexi]=nl;
/* protocol of the position of the new                          */
/* non-empty bin (in frequ[] and index[])                       */
    }
    else
/* if the actual observation is not the first in this bin       */
    {
      frequ[bin[indexi]]++;
/* increase the number of obs. in belonging non-empty bin       */
    }
  }
/* The vector frequ contains nonzero values in positions        */
/* 1,2,3,..,nl.  The other values are prespecified      */
/* as 0, since frequ is installed as a 0-vector.  In            */
/* analogy index contains the indices of the non-empty bins.    */
  for(M=rangem[0];M<rangem[1]+1;M++)
  {
    h=(double)M*delta[0];
/* resetting the vector for the density estimates               */
    for(z=0;z<numbin[0];z++)
      fm[z]=0.0;
/* resetting the estimate for expectation of fm[x]              */
    estexp=0.0;
/* resetting the value of the integral over fm[x]^2             */
    intsquare=0.0;
/* Calculation of the kernel depending weights.                 */
/* The factor c(M) assures that sum(w.M(i/M),1-M,M-1)=M.        */
/* Note, in formula of cm the division by [n * h] does not      */
/* belong to the theoretical value of cm, but is a factor,      */
/* which has to be multiplied one time, since fm=1/(nh)*..      */

    switch ( *kernel )
    {
      case 1 :
/* uniform-kernel K(u)=c(M) * I(|u|<=1)                         */
/* c(M)=M/(2M-1)                                                */
      {
	cm=(double)M/((double)(2*M-1)*(double)n[0]*h);
	for (i=0;i<M;i++)
	  kwe[i]=cm;
	break;
      }
      case 2 :
/* Triangle - kernel K(u)=c(M)*(1- |u|) * I(|u|<=1)             */
/* original Average-Shifted-Histogram-kernel                    */
/* c(M)=1                                                       */
      {
	cm=1.0/((double)n[0]*h);
	for (i=0;i<M;i++)
	  kwe[i]=cm*(1-(double)i/(double)M);
	break;
      }
      case 3 :
/* Epanechnikov - kernel K(u)=c(M) * (1- u*u) * I(|u|<=1)       */
/* c(M)=3*M^2/4*M^2-1)                                          */
      {
	cm=(double) (3*M*M)/((double)(4*M*M-1)
					*(double)n[0]*h);
	for (i=0;i<M;i++)
	  kwe[i]=cm*(1-pow((double)i/(double)M, 2.0));
	break;
      }
      case 4 :
/* Quartic - kernel K(u)= c(M) * (1- u*u)^2 * I(|u|<=1)         */
/* c(M)=15*M^4/(16*M^4-1)                                       */
      {
	cm=0.9375/((1.0-0.0625*pow((double)M,-4.0))
					*(double)n[0]*h);
	for (i=0;i<M;i++)
	  kwe[i]=cm*pow(1-pow((double)i/(double)M,2.0),2.0);
	break;
      }
      case 5 :
/* Triweight-kernel K(u)=c(M)*(1-u*U)^3*I(|u|<=1)               */
/* c(M)=35*M^6/(32*M^6+14/3*M^2-5/3)                            */
      {
	cm=1.09375/((1.0+0.14583333*pow((double)M,-4.0)
	-0.052083333*pow((double)M,-6.0))*(double)n[0]*h);
	for (i=0;i<M;i++)
	  kwe[i]=cm*pow(1-pow((double)i/(double)M,2.0),3.0);
	break;
      }
      default: printf("error in the choice of kernel !");
      return;
    }
/* end of the switch-part                                       */

/* Like Kernel estimates, where a kernel function is set        */
/* around the observations, we group the calculated weights     */
/* around the non-empty bins.                                   */

    for (k=1;k<nl+1;k++)
    {
      for (i=1-M;i<M;i++)
      {
	fm[i+index[k]]=fm[i+index[k]]+kwe[abs(i)]*(double)frequ[k];
      }
    }
  
/* Estimate E[fm(X)] basing on leave one out estimates          */
/* Need only non-empty bins                                     */
    for(k=1;k<nl+1;k++)
      estexp+=fm[index[k]]*(double)frequ[k];
    estexp-=kwe[0]*n[0];
/* Note, weights kwe[] are divided by n*h                       */
    estexp /=(double)(n[0]-1);
/* Calculation of the integral over fm[x]^2                     */
    for(z=0;z<numbin[0];z++)
      intsquare+=fm[z]*fm[z];
    intsquare*=delta[0];
/* Calculation of value of cross-validation basing on           */
/* calculated quantities.                                       */
    cv[M-rangem[0]]=intsquare-2.0*estexp;
  }  /* end of loop over M */
    
}/* end */
   
					
Hbinning(x,y,start,delta,bin,nl,counts,ysum,index,indexobs,n)
double *x,*y,*start,*delta,*ysum;
long *bin,*nl,*counts,*index,*indexobs,*n;
{
  long register i,j;
/* BINNING THE DATA FOR GOLDEN SECTION BOOTSTRAP */
  for (i=0;i<(*n);i++)
  {
    j=floor((x[i]-(*start))/(*delta));
    if (bin[j]==0)
    {
      (*nl)++;
      counts[(*nl)]=1;
      ysum[(*nl)]=y[i];
      index[(*nl)]=j;
      bin[j]=(*nl);
    }
    else
    {
      counts[bin[j]]++;
      ysum[bin[j]]+=y[i];
    }
    indexobs[i]=j;
  }
} /* END FUNCTION */
createweights(kwe,M,kernel)
double *kwe;
long *M,*kernel;
{
  long register i;
/* CREATING WEIGHTS */
  switch (*kernel)
  {
    case 1 :
/* uniform-kernel                                               */
    {
      for (i=0;i<M[0];i++)
	kwe[i]=1;
      break;
    }
    case 2 :
/* Triangle - kernel                                            */
    {
      for (i=0;i<M[0];i++)
	kwe[i]=1.0-(double)i/(double)M[0];
      break;
    }
    case 3 :
/* Epanechnikov - kernel                                        */
    {
      for (i=0;i<M[0];i++)
	kwe[i]=1.0-pow((double)i/(double)M[0], 2.0);
      break;
    }
    case 4 :
/* Quartic - kernel                                             */
    {
      for (i=0;i<M[0];i++)
	kwe[i]=pow(1.0-pow((double)i/(double)M[0],2.0),2.0);
      break;
    }
    case 5 :
/* Triweight-kernel                                             */
    {
      for (i=0;i<M[0];i++)
	kwe[i]=pow(1.0-pow((double)i/(double)M[0],2.0),3.0);
      break;
    }
    default: 
    {  
      printf("error in the choice of kernel %d!", *kernel);
      return;
    }
  }
} /* END FUNCTION */
weightingbins(rM,fM,nl,counts,ysum,index,kwe,M)
double *rM,*fM,*ysum,*kwe;
long *nl,*counts,*index,*M;
{
  long register i,j;
  for(i=1;i<(*nl)+1;i++)
  {
    for (j=1-(*M);j<(*M);j++)
    {
      fM[index[i]+j]+=kwe[abs(j)]*counts[i];
      rM[index[i]+j]+=kwe[abs(j)]*ysum[i];
    }
  }
}/* END FUNCTION */
residuals(res,y,mM,indexobs,n)
double *res,*y,*mM;
long *indexobs,*n;
{
  long register i;
  for(i=0;i<(*n);i++)
  {
    res[i] = y[i] - mM[indexobs[i]];
  }
} /* END FUNCTION */
bootstrapsums(residual,mM,bin,indexobs,ysumstar,n,samples,nl)
double *residual,*mM,*ysumstar;
long *bin,*indexobs,*n,*samples,*nl;
{
  long register i,j;
  long counter,counter2;
  for(i=0;i<(*samples);i++)
  {
    counter=i*(*nl);
    counter2=i*(*n);
    for (j=0;j<(*n);j++)
    {
      ysumstar[counter+bin[indexobs[j]]]+=
		residual[counter2+j]+mM[indexobs[j]];
    }
  }
} /* END FUNCTION */
bootstrapestimate(mMstar,rMstar,fM,nl,ysumstar,kwe,M,
		locindex,locationnumber,bin,samples)
double *mMstar,*rMstar,*fM,*ysumstar,*kwe;
long *nl,*M,*locindex,*locationnumber,*bin,*samples;
{
  long register i,j,k;
  long index1,index2;
  double weight;
  for (j=0;j<(*locationnumber);j++)
  {
    index1=j*(*samples);
    for(k=1-(*M);k<(*M);k++)
    {
      index2=bin[locindex[j]+k];
      if(index2>0)
      {
	weight=kwe[abs(k)];
	for(i=0;i<(*samples);i++)
	{
	  rMstar[index1+i]+=weight*ysumstar[i*(*nl)+index2];
	}
      }
    }
    for(i=0;i<(*samples);i++)
    {
      mMstar[index1+i]=rMstar[index1+i]/fM[locindex[j]];
    }
  }
} /* END FUNCTION */
/* C-PROGRAM FOR KERNEL DENSITY ESTIMATION                      */
/* INTERFACE :                                                  */
/* data = vector of observations                                */
/* n = number of observations                                   */
/* grid = vector with gridpoints                                */
/* fh = vector with estimates                                   */
/* kernel = number of kerneltype between 1 and 6                */
/* 1 = Uniform, 2 = Triangle,                                   */
/* 3 = Epanechnikov, 4 = Quartic,                               */
/* 5 = Triweight, 6 = Gaussian,                                 */
/* 7 = Cosinus                                                  */
/* h = bandwidth of the kernel estimation                       */
/* points = number of gridpoints where the                      */
/* estimation should be calculated                              */
/* range = minimum and maximum of data                          */

#undef PI
#define PI 3.14159265

double kernl1 (u)
double u;
{
  return(0.5);
}

double kernl2 (u)
double u;
{
  double w;
  w = 1-fabs(u);
  return(w);
}

double kernl3 (u)
double u;
{
  double w;
  w = 0.75 *(1-u*u);
  return(w);
}

double kernl4 (u)
double u;
{
  double w1, w2;
  w2 = 1-u*u;
  w1 = 0.9375 *w2 * w2;
  return(w1);
}

double kernl5 (u)
double u;
{
  double w1, w2;
  w2=1-u*u;
  w1 = 1.09375 *w2*w2*w2;
  return(w1);
}

double kernl6 (u)
double u;
{
  double w;
  w = 0.3989*exp(-0.5 *u*u);
  return(w);
}

double kernl7 (u)
double u;
{
  double w;
  w = PI *0.25*cos(0.5* PI * u);
  return(w);
}



kde(data,n,grid,fh,kernel,h,points,range)
double *data,*grid,*fh,*h,*range;
long *n,*kernel,*points;
{
  double gridmin,step,nh;
  long i,j,k1,k2;
  double (*kernelproc)();
  nh=h[0]*(double)n[0];

  switch (kernel[0])
  {
    case 1 :  kernelproc = kernl1; break;
    case 2 :  kernelproc = kernl2; break;
    case 3 :  kernelproc = kernl3; break;
    case 4 :  kernelproc = kernl4; break;
    case 5 :  kernelproc = kernl5; break;
    case 6 :  kernelproc = kernl6; break;
    case 7 :  kernelproc = kernl7; break;
    default:  printf(" Error in the choice of kernel"); return;
  }

/* left boundary of grid                                        */

  gridmin=range[0]-h[0];
  
/* interval between two following points on the grid            */
  step=(2.0*h[0]+range[1]-range[0])/(double)(points[0]-1);

/* grid                                                         */
  for (j=0;j<points[0];j++)
  {
    grid[j]=gridmin+step*(double)j;
  }
  
/* now the kernel function is set around each gridpoint         */

  if(kernel[0]==6)
  {
    for(i=0;i<n[0];i++)
    {
      for(j=0;j<points[0];j++)
      {
	fh[j]+=kernelproc((grid[j]-data[i])/h[0])/nh;
      }
    }
  }
  else
/* Except for the Gaussian kernel (kernel[0]=6)                 */
/* the kernel weights are only positive in a                    */
/* interval fixed by the indices k1 and k2.                     */
  {
    for (i=0;i<n[0];i++)
    {
      k1=ceil((data[i]-h[0]-gridmin)/step);
      k2=floor((data[i]+h[0]-gridmin)/step);
      for (j=k1;j<k2+1;j++)
      {
	fh[j]+=kernelproc((grid[j]-data[i])/h[0])/nh;
      } /* endj */
    }  /* endi */
  } /* endif */
  return;
} /* end */

/* C-PROGRAM FOR KERNEL DENSITY ESTIMATION                      */
/* INTERFACE : fun = kernel function                            */
/* data = vector of observations                                */
/* n = number of observations                                   */
/* grid = vector with gridpoints                                */
/* fh = vector with estimates                                   */
/* kernel = number of kerneltype between 1 and 6                */
/* 1 = Uniform, 2 = Triangle,                                   */
/* 3 = Epanechnikov, 4 = Quartic,                               */
/* 5 = Triweight, 6 = Gaussian,                                 */
/* 7 = Cosinus                                                  */
/* h = bandwidth of the kernel estimation                       */
/* points = number of gridpoints where the                      */
/* estimation should be calculated                              */
/* range = minimum and maximum of data                          */


static void *func;
/* pointer for the S function                                   */
static double sfun();  /* added to make this compile */

kernelestimation(fun,data,n,grid,fh,kernel,h,points,range)
void **fun;
double *data,*grid,*fh,*h,*range;
long *n,*kernel,*points;
/* note the use of void **, since we have sent a list to C      */
{
  double gridmin,step,nh;
  long i,j,k1,k2;
  nh=h[0]*(double)n[0];
  func=fun[0];
  
/* left boundary of grid                                        */

  gridmin=range[0]-h[0];
  
/* interval between two following points on the grid            */
  step=(2.0*h[0]+range[1]-range[0])/(double)(points[0]-1);

/* grid                                                         */
  for (j=0;j<points[0];j++)
  {
    grid[j]=gridmin+step*(double)j;
  }
  
/* now the kernel function is set around each gridpoint         */

  if(kernel[0]==6)
  {
    for(i=0;i<n[0];i++)
    {
      for(j=0;j<points[0];j++)
      {
	fh[j]+=sfun((grid[j]-data[i])/h[0])/nh;
      }
    }
  }
  else
/* Except for the Gaussian kernel (kernel[0]=6)                 */
/* the kernel weights are only positive in a                    */
/* interval fixed by the indices k1 and k2.                     */
  {
    for (i=0;i<n[0];i++)
    {
      k1=ceil((data[i]-h[0]-gridmin)/step);
      k2=floor((data[i]+h[0]-gridmin)/step);
      /*printf("%d %d\n", k1,k2);*/
      for (j=k1;j<k2+1;j++)
      {
     /* printf("%f %f %f %f\n",grid[j],data[i],h[0],(grid[j]-data[j])/h[0]);*/
	fh[j]+=sfun((grid[j]-data[i])/h[0])/nh;
      } /* endj */
    }  /* endi */
  } /* endif */
  return;
} /* end */

/* The following part contains the call of our S-function       */
static double sfun(z)
double z;
{
  void *args[1],*values[1];
  double zz[1], *result;
  char *mode[1];
  long length[1];
  mode[0]="double";
  length[0]=1;
  args[0]=(void *)(zz);
  zz[0]=z;
  call_S(func, 1L, args,mode,length,0,1L,values);
  result=(double *)values[0];
  return(result[0]);
}  
  
kerneldensityci(data,n,h,g,fh,f2g,x,nx)
double *data, *h, *g, *fh, *f2g, *x;
long *n, *nx;
{
  double eval,evalh,evalg,nh,ng,pi;
  long i,j;
  pi=4.0*atan(1.0);
  nh=h[0]*(double)n[0];
  ng=pow(g[0],3.0)*(double)n[0];
  for(j=0;j<nx[0];j++)
  {
    for(i=0;i<n[0];i++)
    {
      eval=x[j]-data[i];
      evalh=pow(eval/h[0],2.0);
      evalg=pow(eval/g[0],2.0);
      fh[j]+=exp(-0.5*evalh)/(nh*sqrt(2.0*pi));
      f2g[j]+=(evalg-1)*exp(-0.5*evalg)/(ng*sqrt(2.0*pi));
    }
  }
}
/* C-PROGRAM FOR KERNEL REGRESSION                              */
/* INTERFACE : x,y vectors of data                              */
/* h = bandwidth of the kernel estimation                       */
/* grid = vector with gridpoints                                */
/* fh = vector with estimates                                   */
/* rh = vector with ??                                          */
/* mh = vector with ??                                          */
/* n = number of observations                                   */
/* gridpoints = number of gridpoints where the                  */
/* estimation should be calculated                              */
/* kernel = number of kerneltype between 1 and 7                */
/* 1 = Uniform, 2 = Triangle,                                   */
/* 3 = Epanechnikov, 4 = Quartic,                               */
/* 5 = Triweight, 6 = Gaussian,                                 */
/* 7 = Cosinus                                                  */
/* rangex = minimum and maximum of data                         */
/* fh0 how to handle case rh/fh=0/0 (default missing value)     */


double kernel1 (u)
double u;
{
  return(0.5);
}

double kernel2 (u)
double u;
{
  double w;
  w = 1-fabs(u);
  return(w);
}

double kernel3 (u)
double u;
{
  double w;
  w = 0.75 *(1-u*u);
  return(w);
}

double kernel4 (u)
double u;
{
  double w1, w2;
  w2 = 1-u*u;
  w1 = 0.9375 *w2 * w2;
  return(w1);
}

double kernel5 (u)
double u;
{
  double w1, w2;
  w2=1-u*u;
  w1 = 1.09375 *w2*w2*w2;
  return(w1);
}

double kernel6 (u)
double u;
{
  double w;
  w = (1.0/pow(2.0*PI,0.5))*exp(-0.5 *u*u);
  return(w);
}

double kernel7 (u)
double u;
{
  double w;
  w = PI / 4.0*cos(PI/2.0 * u);
  return(w);
}


kernelregression(x,y,h,grid,fh,rh,mh,n,gridpoints,kernel,rangex,fh0)
double *x,*y,*grid,*fh,*rh,*mh,*h,*rangex;
long *n,*kernel,*gridpoints,*fh0;
{
  double gridmin,gridstep,weight,distance,estimate;
  long k1,k2;
  long register i,j;
  double (*kernelproc)();
  gridmin=rangex[0]-h[0];
  gridstep=(rangex[1]+h[0]-gridmin)/(double)(gridpoints[0]-1);
  for (i=0;i<gridpoints[0];i++)
    grid[i]=gridmin+gridstep*(double)i;
  switch (kernel[0])
  {
    case 1 :  kernelproc = kernel1; break;
    case 2 :  kernelproc = kernel2; break;
    case 3 :  kernelproc = kernel3; break;
    case 4 :  kernelproc = kernel4; break;
    case 5 :  kernelproc = kernel5; break;
    case 6 :  kernelproc = kernel6; break;
    case 7 :  kernelproc = kernel7; break;
    default:  printf(" Error in the choice of kernel"); return;
  }
  if(kernel[0]!=6)
/* If the kernel is not Gaussian, i.e. has support [-1,1]       */
  {
    for (i=0;i<n[0];i++)
    {
      k1=floor((x[i]-grid[0]-h[0])/gridstep)+1;
      k2=floor((x[i]-grid[0]+h[0])/gridstep);
      for (j=k1;j<(k2+1);j++)
      {
	distance=(x[i]-grid[j])/h[0];
	weight=(*kernelproc)(distance);
	rh[j]+=weight*y[i];
	fh[j]+=weight;
      } 
    } 
  }
  else 
  {
    for(i=0;i<n[0];i++)
    {
      for(j=0;j<gridpoints[0];j++)
      {
	distance=(x[i]-grid[j])/h[0];
	weight=(*kernelproc)(distance);
	rh[j]+=weight*y[i];
	fh[j]+=weight;
      }
    }
  }
  if((*fh0)==1)
  {
/* Estimate is 0, if fh[j]=0 and flag (*fh0) is set to 1        */
    for(j=0;j<gridpoints[0];j++)
    {
      if(fh[j]==0)
	mh[j]=0;
      else
	mh[j]=rh[j]/fh[j];
    }
  }
  else
  {
    for(j=0;j<gridpoints[0];j++)
      mh[j]=rh[j]/fh[j];
  }
  return;
} /* end */

  
/* C-routine of WARPing-density-estimation for                  */
/* 1-dimensional data with 6 different kernels                  */

warp(x,fh,h,wM,origin,binmesh,counts,J,M,n,kernel)

/* x <==> data                                                  */
/* fh <==> vector == 0 will contain density estimates           */
/* h bandwidth                                                  */
/* wM <==> weights depending on the differnet kernels           */
/* origin <==> left border of first bin                         */
/* binmesh <==> vector == 0 protocol for non-empty bins         */
/* counts <==> counts of observations in non-empty bins         */
/* J <==> index of non-empty bins                               */
/* M <==> number of small bins in one large bin                 */
/* n <==> number of data                                        */
/* kernel <==> type of kernel coded with 1 to 5                 */

double*x,*fh,*h,*wM,*origin;
long*M,*n,*kernel,*binmesh,*counts,*J;
{
  register long i,k,index;
  double delta,cm;
  long binnumber;
  delta=h[0]/(double)M[0];
  binnumber=0;
  
/* delta binwidth of small bins <==> h/M                        */
/* binnumber, integer to count the non-empty bins               */

/* Binning the data                                             */

  for (i=0;i<n[0];i++)
  {
    index=floor((x[i]-origin[0])/delta);
/* the actual obs. belongs to the bin with number index         */
    if (binmesh[index]==0)
/* if the actual observation is the first in that bin           */
    {
      binnumber++;
/* the number of non-empty bins increases by 1                  */
      counts[binnumber] = 1;
/* the number of obs. in this bin is 1 (at this time)           */
      J[binnumber]=index;
      binmesh[index]=binnumber;
/* protocol of the position of the new                          */
/* non-empty bin (in counts[] and J[])                          */
    }
    else
/* if the actual observation is not the first in this bin       */
    {
      counts[binmesh[index]]++;
/* increase the number of obs. in belonging non-empty bin       */
    }
  }
/* The vector counts contains nonzero values in positions       */
/* 1,2,3,..,binnumber.  The other values are prespecified       */
/* as 0, since counts is installed as a 0-vector.  In           */
/* analogy J contains the indices of the non-empty bins.        */

/* Calculation of the kernel depending weights.                 */
/* The factor c(M) assures that sum(w.M(i/M),1-M,M-1)=M.        */
/* Note, in formula of cm the division by [n * h] does not      */
/* belong to the theoretical value of cm, but is a factor,      */
/* which has to be multiplied one time, since fh=1/(nh)*..      */

  switch ( *kernel )
  {
    case 1 :
/* uniform-kernel K(u)=c(M) * I(|u|<=1)                         */
/* c(M)=M/(2M-1)                                                */
    {
      cm=(double)M[0]/((double)(2*M[0]-1)*(double)n[0]*h[0]);
      for (i=0;i<M[0];i++)
	wM[i]=cm;
      break;
    }
    case 2 :
/* Triangle - kernel K(u)=c(M)*(1- |u|) * I(|u|<=1)             */
/* original Average-Shifted-Histogram-kernel                    */
/* c(M)=1                                                       */
    {
      cm=1.0/((double)n[0]*h[0]);
      for (i=0;i<M[0];i++)
	wM[i]=cm*(1-(double)i/(double)M[0]);
      break;
    }
    case 3 :
/* Epanechnikov - kernel K(u)=c(M) * (1- u*u) * I(|u|<=1)       */
/* c(M)=3*M^2/4*M^2-1)                                          */
    {
      cm=(double) (3*M[0]*M[0])/((double)(4*M[0]*M[0]-1)
					*(double)n[0]*h[0]);
      for (i=0;i<M[0];i++)
	wM[i]=cm*(1-pow((double)i/(double)M[0], 2.0));
      break;
    }
    case 4 :
/* Quartic - kernel K(u)= c(M) * (1- u*u)^2 * I(|u|<=1)         */
/* c(M)=15*M^4/(16*M^4-1)                                       */
    {
      cm=0.9375/((1.0-0.0625*pow((double)M[0],-4.0))
					*(double)n[0]*h[0]);
      for (i=0;i<M[0];i++)
	wM[i]=cm*pow(1-pow((double)i/(double)M[0],2.0),2.0);
      break;
    }
    case 5 :
/* Triweight-kernel K(u)=c(M)*(1-u*U)^3*I(|u|<=1)               */
/* c(M)=35*M^6/(32*M^6+14/3*M^2-5/3)                            */
    {
      cm=1.09375/((1.0+0.14583333*pow((double)M[0],-4.0)
      -0.052083333*pow((double)M[0],-6.0))*(double)n[0]*h[0]);
      for (i=0;i<M[0];i++)
	wM[i]=cm*pow(1-pow((double)i/(double)M[0],2.0),3.0);
      break;
    }
    case 6 :
/* Gaussian kernel */
    {
      cm = 0.3989*4.0 /((double)n[0]*h[0]);
      for (i=0;i<M[0];i++)
	wM[i]=cm*exp(-8.0*pow((double)i/(double)M[0],2.0));
      break;
    }
    default: printf("error in the choice of kernel !");
    return;
  }
/* end of the switch-part                                       */

/* Like Kernel estimates, where a kernel function is set        */
/* around the observations, we group the calculated weights     */
/* around the non-empty bins.                                   */

  for (k=1;k<binnumber+1;k++)
  {
    for (i=1-M[0];i<M[0];i++)
    {
      fh[i+J[k]]=fh[i+J[k]]+wM[abs(i)]*(double)counts[k];
    }
  }
}
   
					
/* C-routine of WARPing-density-estimation for                  */
/* 1-dimensional data with 5 different kernels                  */

warping(x,fh,h,wM,origin,binmesh,counts,J,M,n,kernel)

/* x <==> data                                                  */
/* fh <==> vector == 0 will contain density estimates           */
/* h bandwidth                                                  */
/* wM <==> weights depending on the differnet kernels           */
/* origin <==> left border of first bin                         */
/* binmesh <==> vector == 0 protocol for non-empty bins         */
/* counts <==> counts of observations in non-empty bins         */
/* J <==> index of non-empty bins                               */
/* M <==> number of small bins in one large bin                 */
/* n <==> number of data                                        */
/* kernel <==> type of kernel coded with 1 to 5                 */

double*x,*fh,*h,*wM,*origin;
long*M,*n,*kernel,*binmesh,*counts,*J;
{
  register long i,k,index;
  double delta,cm;
  long binnumber;
  delta=h[0]/(double)M[0];
  binnumber=0;
  
/* delta binwidth of small bins <==> h/M                        */
/* binnumber, integer to count the non-empty bins               */

/* Binning the data                                             */

  for (i=0;i<n[0];i++)
  {
    index=floor((x[i]-origin[0])/delta);
/* the actual obs. belongs to the bin with number index         */
    if (binmesh[index]==0)
/* if the actual observation is the first in that bin           */
    {
      binnumber++;
/* the number of non-empty bins increases by 1                  */
      counts[binnumber] = 1;
/* the number of obs. in this bin is 1 (at this time)           */
      J[binnumber]=index;
      binmesh[index]=binnumber;
/* protocol of the position of the new                          */
/* non-empty bin (in counts[] and J[])                          */
    }
    else
/* if the actual observation is not the first in this bin       */
    {
      counts[binmesh[index]]++;
/* increase the number of obs. in belonging non-empty bin       */
    }
  }
/* The vector counts contains nonzero values in positions       */
/* 1,2,3,..,binnumber.  The other values are prespecified       */
/* as 0, since counts is installed as a 0-vector.  In           */
/* analogy J contains the indices of the non-empty bins.        */

/* Calculation of the kernel depending weights.                 */
/* The factor c(M) assures that sum(w.M(i/M),1-M,M-1)=M.        */
/* Note, in formula of cm the division by [n * h] does not      */
/* belong to the theoretical value of cm, but is a factor,      */
/* which has to be multiplied one time, since fh=1/(nh)*..      */

  switch ( *kernel )
  {
    case 1 :
/* uniform-kernel K(u)=c(M) * I(|u|<=1)                         */
/* c(M)=M/(2M-1)                                                */
    {
      cm=(double)M[0]/((double)(2*M[0]-1)*(double)n[0]*h[0]);
      for (i=0;i<M[0];i++)
	wM[i]=cm;
      break;
    }
    case 2 :
/* Triangle - kernel K(u)=c(M)*(1- |u|) * I(|u|<=1)             */
/* original Average-Shifted-Histogram-kernel                    */
/* c(M)=1                                                       */
    {
      cm=1.0/((double)n[0]*h[0]);
      for (i=0;i<M[0];i++)
	wM[i]=cm*(1-(double)i/(double)M[0]);
      break;
    }
    case 3 :
/* Epanechnikov - kernel K(u)=c(M) * (1- u*u) * I(|u|<=1)       */
/* c(M)=3*M^2/4*M^2-1)                                          */
    {
      cm=(double) (3*M[0]*M[0])/((double)(4*M[0]*M[0]-1)
					*(double)n[0]*h[0]);
      for (i=0;i<M[0];i++)
	wM[i]=cm*(1-pow((double)i/(double)M[0], 2.0));
      break;
    }
    case 4 :
/* Quartic - kernel K(u)= c(M) * (1- u*u)^2 * I(|u|<=1)         */
/* c(M)=15*M^4/(16*M^4-1)                                       */
    {
      cm=0.9375/((1.0-0.0625*pow((double)M[0],-4.0))
					*(double)n[0]*h[0]);
      for (i=0;i<M[0];i++)
	wM[i]=cm*pow(1-pow((double)i/(double)M[0],2.0),2.0);
      break;
    }
    case 5 :
/* Triweight-kernel K(u)=c(M)*(1-u*U)^3*I(|u|<=1)               */
/* c(M)=35*M^6/(32*M^6+14/3*M^2-5/3)                            */
    {
      cm=1.09375/((1.0+0.14583333*pow((double)M[0],-4.0)
      -0.052083333*pow((double)M[0],-6.0))*(double)n[0]*h[0]);
      for (i=0;i<M[0];i++)
	wM[i]=cm*pow(1-pow((double)i/(double)M[0],2.0),3.0);
      break;
    }
    default: printf("error in the choice of kernel !");
    return;
  }
/* end of the switch-part                                       */

/* Like Kernel estimates, where a kernel function is set        */
/* around the observations, we group the calculated weights     */
/* around the non-empty bins.                                   */

  for (k=1;k<binnumber+1;k++)
  {
    for (i=1-M[0];i<M[0];i++)
    {
      fh[i+J[k]]=fh[i+J[k]]+wM[abs(i)]*(double)counts[k];
    }
  }
}
   
					
/* Algorithm for calculation of the WARPing-approximation       */
/* of NADARAYA-WATSON -regression estimate                      */
/* 1-dimensional predictor with 5 different kernels             */

warpingregression(x,y,h,M,fM,rM,mM,kwe,origin,bin,frequ,ysum,
			index,n,numbin,kernel,fM0)

/* x <==> 1 - dim predictor                                     */
/* y <==> 1 - dim response                                      */
/* h <==> smoothing parameter                                   */
/* M <==> number of small bin in one large window               */
/* fM <==> density estimation of predictor density              */
/* rM <==> estimation of numerator                              */
/* mM <==> conditional expectation of y given x                 */
/* kwe <==> weights depending on the different kernels          */
/* origin <==> left boundary of first bin                       */
/* bin <==> flag for non-empty bins                             */
/* frequ <==> counts  in non-empty bins                         */
/* ysum <==> sum of response-observations in non-empty bins     */
/* index <==> index of non-empty bins                           */
/* n <==> number of observations                                */
/* numbin <==> total number of small bins                       */
/* kernel <==> type of kernel coded 1 to 5                      */
/* fM0 <==> flag for na-handling                                */

double*x,*y,*fM,*rM,*mM,*h,*kwe,*origin,*ysum;
long*M,*n,*kernel,*bin,*frequ,*index,*numbin,*fM0;
{
  long i,j,k;
  double delta, tmp;
  long iabs,indexi,nl;

  delta=h[0]/(double)M[0];
  nl=0;
/* Binning the data                                             */

  for (i=0;i<n[0];i++)
  {
    j=floor((x[i]-origin[0])/delta);
    if (bin[j]==0) /* first observation in bin j */
    {
      ++nl;
      frequ[nl] = 1;
      ysum[nl]=y[i];
      index[nl]=j;
      bin[j]=nl;
    }
    else
    {
      frequ[bin[j]]++;
      ysum[bin[j]]+=y[i];
    }
  }

/* Creating weights.                                            */
/* Note. weights are not normalised                             */
  switch ( *kernel )
  {
    case 1 :
/* uniform-kernel                                               */
    {
      for (i=0;i<M[0];i++)
	kwe[i]=1;
      break;
    }
    case 2 :
/* Triangle - kernel                                            */
    {
      for (i=0;i<M[0];i++)
	kwe[i]=1.0-(double)i/(double)M[0];
      break;
    }
    case 3 :
/* Epanechnikov - kernel                                        */
    {
      for (i=0;i<M[0];i++)
	kwe[i]=1.0-pow((double)i/(double)M[0], 2.0);
      break;
    }
    case 4 :
/* Quartic - kernel                                             */
    {
      for (i=0;i<M[0];i++) {
	tmp = (double)i/(double)M[0];
	tmp = tmp*tmp;
	kwe[i]=(1.0-tmp)*(1.0-tmp);
      }      
      break;
    }
    case 5 :
/* Triweight-kernel                                             */
    {
      for (i=0;i<M[0];i++)
	kwe[i]=pow(1.0-pow((double)i/(double)M[0],2.0),3.0);
      break;
    }
    default: 
    {  
      printf("error in the choice of kernel %d!", *kernel);
      return;
    }
  }
/* end of the switch-part                                       */

/* Weighting the bins                                           */
  for (k=1;k<nl+1;k++)
  {
    for (i=1-M[0];i<M[0];i++)
    {
      indexi=i+index[k];
      iabs=abs(i);
      fM[indexi]+=kwe[iabs]*(double)frequ[k];
      rM[indexi]+=kwe[iabs]*ysum[k];
    }
  }
  
/* combination of numerator and denominator                     */
  if((*fM0)==1)
  {
/* Estimate is set to 0, if fM[j]=0                             */
    for(j=0;j<(*numbin);j++)
    {
      if(fM[j]==0)
	mM[j]=0;
      else
	mM[j]=rM[j]/fM[j];
    }
  }
  else
  {
/* generation of missings possible, if fM[j]=0                  */
    for(j=0;j<(*numbin);j++)
      mM[j]=rM[j]/fM[j];
  }
} /* end */
   
